#ifndef _typing_h_
#define _typing_h_
#include <stdio.h> /* FILE */
#include "main.h"

/* magic union name for exceptions as passed from phamc */
#define EXCEPTIONUNION "#exception"
#define LISTUNION "list"


/*
 * Type are often passed as strings:
 *
 * <type> :=
 *   'c'                               - 8-bit character
 *   'b'                               - boolean
 *   'i' <size>                        - 'size'-bit integer
 *   'w' <size>                        - 'size'-bit word  - often aliases to 'i' <size>
 *   'f' <size>                        - 'size'-bit float
 *   's' <size>                        - string of possibly bounded length 'size'
 *   'v' <size> <type>                 - vector (of possibly bounded 'size') of 'type'
 *   'l' <size> <type>                 - ['type'], possibly of bounded size - often aliases to '(list:' <type> ')'
 *   't' <size> <type1> .. <type_size> - tuple
 *   '(' <name> ':' <type>* ')'        - union type 'name' as parametrised by the <type>*
 *   'a' <idx>                         - reference within paramertised union type
 *
 * <size> := ['?' | '0' | [0..9]*]     - note: optional in all cases
 *
 * exceptions are treated as unions - but unions that are always tagged structures
 */


typedef enum {TINT, TFLOAT, TWORD, TCHAR, TBOOL, TSTRING, TLIST, TVECT, TTUPLE, TUNION, TARG} type_e;

struct type_t; typedef struct type_t type_t;
struct union_t; typedef struct union_t union_t;
struct unionc_t; typedef struct unionc_t unionc_t;


struct type_t {
	type_e tag;
	int size;   /* sub size for union/tuple, otherwise is just precision */
	union_t *u; /* only when tag==TUNION */
	type_t *sub[];
};

type_t *allocType(type_e tag, int size);
void freeType(type_t *t);
boolean eqType(type_t *x, type_t *y);
type_t *cloneType(type_t *t);
boolean simpleType(type_t *t);


/* string -> type */
type_t *parseType(const char **p);

/* type -> string (no buffer overflow checks) */
void showType(type_t *t, char **p);


/* 
 * Code generation based on types 
 * - all follows a similar scheme
 */
struct tlist_t; typedef struct tlist_t tlist_t;
struct gen_t; typedef struct gen_t gen_t;

struct tlist_t {
	tlist_t *next;
        type_t *type;
        int index;
};

struct gen_t {
	tlist_t *map;                              /* types -> index */
	void(*out)(FILE *, type_t *, int, int []); /* for code gen */
	type_t *(*norm)(type_t *t);                /* simplifying the type */
};


/* produces a normalized clone of the type - performs type substituions as necessary */
type_t *normalizeGen(type_t *t, type_t *args[], type_t *(*op)(type_t *t));
void doGen(FILE *file, gen_t *g, type_t *t, int idx);
int lookupGen(FILE *file, gen_t *g, type_t *t);




struct unionc_t {
	const char *name;
	type_t *type;
	union_t *parent; /* convenience */
	boolean builtin;
	unionc_t *next;
};

struct union_t {
	const char *name;
	int arity;
	boolean builtin;
	unionc_t *constrs;
	union_t *next;
};


/*
 * read the HAM instructions and decode the union/type information
 */
void initUnions();

union_t *findUnion(const char *name);
unionc_t *findUnionC(const char *name);

unsigned int enumsInUnion(union_t *u); /* number of zero arg varients */
unsigned int wrapsInUnion(union_t *u); /* number of non-zero arg varients */

/* 
 * non-zero
 * For normal unions is guaranteed to be broken pointers, e.g. 1,5,9,13.. 
 * For exceptions is 1,2,3,4 as these are always wrapped unions, and the enumerations need to be obvious/stable
 */
unsigned int tagForUnionC(unionc_t *c);

boolean isEnumUnionC(unionc_t *c);


/*
 * for unions with a single wrap, that particular wrapped constructor is untagged 
 * - returns 0 if n/a
 */
unsigned int tagForSingleWrap(union_t *u);

/*
 * for debugging - dumps to stdout
 */
void dumpType(type_t *t);
void dumpUnions();

/*
 * utils...
 */
union_t *parseUnion(const char **p);
boolean simpleTypeStr(const char *type); /* convenience that works on type strings */

#endif /*_typing_h_*/
